from datetime import datetime, timedelta
from dateutil.parser import isoparse
import pytz
import pandas as pd
import numpy as np
import logging
from typing import Dict, List, Any, Optional, Union

from arthurai.common.exceptions import UserTypeError, UserValueError, InternalValueError


logger = logging.getLogger(__name__)


"""
Only client facing functions should live in this module. If a util  is used for a specific package and not client facing
it should be added to that packages util.py file.
"""


def format_timestamps(inferences: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Checks list of json inferences to ensure their timestamps have been converted to ISO8601 conventions.

    :param inferences: Input to `send_inferences()` function.

    :return: updated list of json inferences with ISO8601 formatted timestamps
    :raise: TypeError: timestamp is not of type `datetime.datetime`
    :raise: ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """
    # TODO: deprecate support for strings and delete conditional
    #  @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
    if len(inferences) > 0 and (
        is_string_date(inferences[0].get("inference_timestamp"))
        or is_string_date(inferences[0].get("ground_truth_timestamp"))
    ):
        date_to_log = inferences[0].get("inference_timestamp") or inferences[0].get(
            "ground_truth_timestamp"
        )
        logger.warning(
            f"DEPRECATION WARNING: Your provided timestamps starting with: {date_to_log} is of type `str` and is "
            f"assumed to be in UTC format. Timezone information will be ignored. `str` types will be deprecated. "
            f"Please use timeaware `datetime.datetime` timestamps and use `arthurai.util.format_timestamp` if needed, "
            f"e.g. `format_timestamp(datetime.strptime('2020-01-01 01:01:01', '%Y-%m-%d %H:%M:%S'), 'US/Eastern')`."
        )

    for inf in inferences:
        for timestamp_key in ["inference_timestamp", "ground_truth_timestamp"]:
            if timestamp_key in inf:
                inf[timestamp_key] = format_timestamp(inf[timestamp_key])
    return inferences


def format_timestamp(timestamp: Union[datetime, str], location: Optional[str] = None) -> str:
    """Check if timestamp is time aware and convert to ISO8601 conventioned string.

    This helper function converts datetime objects into timezone aware ISO8601 strings, which is necessary when sending
    JSON to Arthur's backend. If `timestamp` argument is timezone aware, no `location` needs to be provided; otherwise,
    a string pytz `location` like "US/Eastern" needs to be provided to establish timezone. String `timestamp` are
    supported for backwards compatability and for simplicity are assumed to already be in UTC format, but string support
    will be deprecated.

    :param timestamp: timestamp to format
    :param location: pytz location as string

    :return: ISO8601 formatted timestamp
    :raise: TypeError: timestamp is not of type `datetime.datetime`
    :raise: ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """

    # TODO: deprecate support for strings and delete conditional
    #  @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
    if isinstance(timestamp, str):
        if timestamp[-1] != "Z":
            return timestamp + "Z"
        return timestamp

    if not (location is None or isinstance(location, str)):
        raise UserTypeError(f"`location` must be valid pytz string like 'US/Eastern' or None type")

    localization = pytz.timezone(location) if location else None

    if not isinstance(timestamp, datetime):
        raise UserTypeError(
            f"`timestamp` {timestamp} must be of type `datetime.datetime`, not {type(timestamp)}."
        )

    if not (
        timestamp.tzinfo is not None
        and timestamp.tzinfo.utcoffset(timestamp) is not None
    ):
        if localization:
            return (
                localization.localize(timestamp).astimezone(pytz.utc).isoformat()[:-6]
                + "Z"
            )  # convert to utc, log warning if converting from non-utc to utc
        else:
            raise UserValueError(
                f"`timestamp` {timestamp} is not timezone aware, please supply pytz `timezone` string to make "
                f"timestamp timezone aware, e.g. `format_timestamp(datetime.datetime(2021, 1, 1), 'US/Eastern')`."
            )
    return timestamp.astimezone(pytz.utc).isoformat()[:-6] + "Z"


# TODO: deprecate support for strings and delete method
#  @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
def is_string_date(string: Any) -> bool:
    """
    Return whether the string can be interpreted as a (partial) ISO-8601 date.

    :param string: str, string to check for date
    """
    if not isinstance(string, str) or string.isdigit():
        return False
    try:
        isoparse(string)
    except ValueError:
        return False
    else:
        return True


def normal_random_ints_fixed_sum(num_values: int, total_sum: int, relative_std_dev: float = 0.5) -> np.ndarray:
    """
    Return `num_values` roughly normally-distributed integers summing to `total_sum`. Numbers are first sampled from
    the normal distribution, then adjusted to fit the sum. Adjustments are made first by shifting the resulting
    distribution to be at least `total_sum`, then decrementing a subset of the values to match `total_sum`.
    :param num_values: number of integers to return
    :param total_sum: total sum the integers returned
    :param relative_std_dev: the relative standard deviation (std_dev / mean) of the initial distribution.
    :return:
    """
    # first pull from the normal distribution to get some initial values
    mean_value = total_sum / num_values
    std_dev = mean_value * relative_std_dev
    initial_values = np.round(np.random.normal(loc=mean_value, scale=std_dev, size=num_values)).astype(np.int64)
    # bump any negatives values up to zero
    positive_values = np.clip(initial_values, a_min=0, a_max=None)

    # shift the distribution to get close to the correct sum
    mean_diff_floor = np.floor((positive_values.sum() - total_sum) / num_values).astype(np.int64)
    values = positive_values - mean_diff_floor

    # now our difference must be less than the length of the array, and positive
    final_difference = values.sum() - total_sum
    # we can simply subtract 1 from however many indices we need to to reach the exact sum
    inds_to_decrement = np.random.choice(num_values, size=final_difference, replace=False)
    values[inds_to_decrement] -= 1

    # failsafe check
    if values.sum() != total_sum:
        raise InternalValueError("Computed random integers do not sum to expected total. Please report to Arthur with "
                                 f"the following debug info: total_sum={total_sum}, num_values={num_values}, "
                                 f"mean_diff_floor={mean_diff_floor}, final_difference={final_difference}, "
                                 f"positive_values={positive_values}, inds_to_decrement={inds_to_decrement}")

    return values


def _parse_duration(time_ref: str) -> timedelta:
    """
    Parses a time duration string. Supports days and hours, e.g. "14d", "-3h", etc.
    :param time_ref: the time string to parse
    :return: timedelta of the parsed value
    """
    raw_value, unit = time_ref[:-1], time_ref[-1:]
    try:
        value = int(raw_value)
    except ValueError as e:
        raise UserValueError(f"Cannot parse time reference value {raw_value}, must be an integer") from e
    if unit == "d":
        return timedelta(days=value)
    elif unit == "h":
        return timedelta(hours=value)
    else:
        raise UserValueError(f"Cannot parse time reference unit {unit}, only days ('d') and hours ('h') are supported")


# for mocking datetime.now without clobbering the entire datetime class
def _datetime_now(*args, **kwargs):
    return datetime.now(*args, **kwargs)


def generate_timestamps(total: int,
                        duration: Union[str, timedelta] = "7d",
                        end: Union[str, datetime] = "now",
                        freq: Union[str, pd.DateOffset] = "D",
                        noisiness: float = 0.3) -> List[datetime]:
    """
    Generates timestamps over a period of time, defaulting to daily timestamps over the last week. Creates `uniques`
    unique timestamp values evenly spaced between `start` and `end` inclusively. Repeats values if
    total ≠ uniques. This is useful for generating timestamps for test inferences outside of a production setting.

    :param total: the total number of timestamps to generate
    :param end: the final timestamp value
    :param duration: the difference between the first and last timestamp values. day or hour strings (e.g. "7d", "4h")
        and timedelta objects are supported
    :param freq: the frequency with which to generate unique values. See `Pandas Offset Aliases
    <https://pandas.pydata.org/docs/user_guide/timeseries.html#offset-aliases>`_
    :param noisiness: how much to noise to add to the number of timestamps repeated at each interval. specifically, the
        coefficient of variation (standard deviation / mean) to use when selecting the number of times to repeat each
        timestamp
    :raises UserValueError: if an input is invalid or values contain future timestamps and allow_future is False
    :return: a list of generated timestamps
    """
    if end == "now":
        end_dt = _datetime_now(pytz.utc)
    elif not isinstance(end, datetime):
        raise UserTypeError(f"'end' should be string 'now' or a datetime object, got type {type(end)} with value {end}")
    else:
        end_dt = end

    if isinstance(duration, str):
        duration = _parse_duration(duration)

    try:
        unique_values = pd.date_range(start=(end_dt - duration), end=end_dt, freq=freq).to_pydatetime()
    except ValueError as e:
        raise UserValueError(e) from e
    unique_count = len(unique_values)

    if unique_count > total:
        raise UserValueError(f"Timestamps with duration {duration} and frequency {freq} yield {unique_count} unique "
                             f"values, but you asked for {total} total values. Total values must be greater than or "
                             "equal to the number of unique timestamps.")

    # determine how many times each timestamp should be repeated
    repeat_counts = normal_random_ints_fixed_sum(unique_count, total, relative_std_dev=noisiness)
    timestamps = unique_values.repeat(repeat_counts).tolist()

    return timestamps



from dataclasses import dataclass
from typing import Union, Optional, Dict, Any

from arthurai.common.constants import ListableStrEnum
from arthurai.core.base import ArthurBaseJsonDataclass, NumberType


class AlertRuleBound(ListableStrEnum):
    Upper = "upper"
    Lower = "lower"


class AlertRuleSeverity(ListableStrEnum):
    Warning = "warning"
    Critical = "critical"


class AlertStatus(ListableStrEnum):
    Resolved = "resolved"
    Acknowledged = "acknowledged"


@dataclass
class AlertRule(ArthurBaseJsonDataclass):
    bound: AlertRuleBound
    threshold: NumberType
    metric_id: str
    severity: AlertRuleSeverity
    name: Optional[str] = None
    lookback_period: Optional[NumberType] = None
    subsequent_alert_wait_time: Optional[NumberType] = None
    enabled: bool = True
    id: Optional[str] = None
    metric_name: Optional[str] = None


@dataclass
class Alert(ArthurBaseJsonDataclass):
    id: str
    timestamp: str
    metric_value: float
    message: str
    model_id: str
    status: str
    alert_rule: AlertRule
    window_start: Optional[str] = None
    window_end: Optional[str] = None
    batch_id: Optional[str] = None


class MetricType(ListableStrEnum):
    ModelOutputMetric = "model_output_metric"
    ModelInputDataMetric = "model_input_data_metric"
    ModelPerformanceMetric = "model_performance_metric"
    ModelDataDriftMetric = "model_data_drift_metric"


@dataclass
class Metric(ArthurBaseJsonDataclass):
    id: str
    name: str
    query: Dict[str, Any]
    is_default: bool
    type: Optional[MetricType] = None
    attribute: Optional[str] = None



import logging
from io import BytesIO
from time import sleep

import requests
from requests.exceptions import RequestException
from typing import Dict, List, Union, Optional, BinaryIO, Tuple, Any
import json as jsonlib
from urllib.parse import urlparse

from requests_toolbelt import MultipartEncoder

from arthurai.client.helper import construct_url
from arthurai.client.http.base import AbstractHTTPClient
from arthurai.client import validation
from arthurai.common.exceptions import InternalTypeError, ResponseServerError, ResponseRedirectError,\
    ArthurInternalError, UserValueError

logger = logging.getLogger(__name__)


def _parse_response(response: requests.Response) -> Union[Dict, List, bytes, BytesIO]:
    """Depending on the type of response from the server, parses the response and returns

    :param response: response from the REST call
    :return: parsed response
    """
    if response is None:
        return bytes()
    # return error codes as raw responses
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
    if response.status_code >= 400:
        return response.content
    if response.request.headers.get('Accept') == 'application/octet-stream':
        return BytesIO(response.content)
    # return the content if the the return type is image/binary/csv
    if response.headers.get('Content-Type') in ('image/jpeg', 'text/csv', 'avro/binary', 'parquet/binary'):
        return response.content

    try:
        return response.json()
    except ValueError:
        return response.content


def _validate_base_url(base_url) -> None:
    """
    Validate that a base URL does not contain any paths or query parameters
    :param base_url: the base URL to validate
    :return: None
    :raises UserValueError: if the URL is invalid
    """
    # add scheme and initial validation from construct_url function
    parsed = urlparse(construct_url(base_url))

    # validate
    if parsed.scheme not in ("http", "https"):
        raise UserValueError(f"Base URL scheme '{parsed.scheme}' is invalid, should be 'http' or 'https'")
    if parsed.path is not None and parsed.path not in ("", "/"):
        raise UserValueError(f"Base URL {base_url} should not contain a path")
    if parsed.query is not None and parsed.query != "":
        raise UserValueError(f"Base URL {base_url} should not contain query parameters")
    if parsed.params is not None and parsed.params != "":
        raise UserValueError(f"Base URL {base_url} should not contain a path parameter")
    if parsed.fragment is not None and parsed.fragment != "":
        raise UserValueError(f"Base URL {base_url} should not contain a fragment")


class HTTPClient(AbstractHTTPClient):
    """
    A `requests`-based HTTP Client intended for interacting with JSON-based APIs.
    Supports response validation, retries, connection reuse, and multipart requests.
    """

    def __init__(self, base_url: str, path_prefix: str = None, default_headers: Dict[str, str] = None,
                 verify_ssl: bool = True, timeout_sec: float = 300.0):
        """
        :param base_url:
        :param path_prefix:
        :param default_headers:
        :param verify_ssl:
        :param timeout_sec:
        """
        self.session = requests.Session()
        if default_headers is not None:
            self.session.headers.update(default_headers)
        self.session.verify = verify_ssl
        self.timeout = timeout_sec

        # exponential backoff settings
        self.BACKOFF_CONSTANT = 0.1
        self.BACKOFF_EXPONENT_BASE = 1.5

        try:
            _validate_base_url(base_url)
        except UserValueError as e:
            raise UserValueError("Base URL is invalid, see nested exception for details. Base URL should not contain a "
                                 "/path or &query=parameters, a /path;parameter, or a #fragment") from e

        if path_prefix is None:
            self.api_base_url = construct_url(base_url, validate=True)
        else:
            self.api_base_url = construct_url(base_url, path_prefix, validate=True)

    def send(self,
             endpoint: str,
             method: str = 'GET',
             json: Optional[Union[Dict, List, str, bytes]] = None,
             files: Optional[Union[Dict[str, BinaryIO], List[Tuple], Dict[str, Tuple]]] = None,
             headers: Optional[Dict[str, str]] = None,
             params: Optional[Union[Dict, bytes]] = None,
             return_raw_response: bool = False,
             retries: int = 0,
             validate_response_status: bool = True,
             validation_response_code: Optional[int] = None) -> Union[Dict, List, bytes, BytesIO, requests.Response]:
        """
        Send an HTTP request
        :param endpoint: the specific endpoint to append to the client URL
        :param method: the HTTP method to use
        :param headers: headers to use for this request in addition to the client default headers
        :param json: data to send as JSON, either a string/bytes to send directly or a dictionary/list to serialize. if
                     `files` is also supplied, this should be a map from name to content, to be sent along with the
                     `files` as a multipart request
        :param files: a map from file names to file-like objects, to be sent as multipart/form-data
        :param params: query parameters to add to the request
        :param return_raw_response: if true, return the requests.Response object received; otherwise attempt to parse
                                    the response
        :param retries: number of times to retry the request on failure. uses exponential backoff
        :param validate_response_status: if True, raise an ArthurException if the status code is not 2XX or does not
            match `validation_response_code`
        :param validation_response_code: expected status code of the response to validate. if None, allow any 2XX
        :return: if return_raw_response is true, return the requests.Response object received; otherwise attempt to
        parse the response
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if headers is None:
            headers = {}

        # VALIDATION AND PREPROCESSING
        if retries < 0:
            raise UserValueError(f"retries must be greater than or equal to 0")
        if validation_response_code is None and retries > 0:
            logger.warning(f"retries was specified as {retries} but validation_response_code was not set, response "
                           f"contents will not be evaluated.")
        # automatically add json content type headers and serialize json if `json` is supplied but no (multipart)
        #  files are supplied
        multipart: bool = files is not None or ('Content-Type' in headers.keys() and
                                                headers['Content-Type'] == "multipart/form-data")
        if json is not None and not multipart:
            logger.debug("serializing JSON and adding Content-Type header")
            # headers
            if 'Content-Type' not in headers.keys():
                headers['Content-Type'] = "application/json"
            elif headers['Content-Type'] != "application/json":
                logger.debug(f"Content-Type header is specified as {headers['Content-Type']}, not overwriting")
            # body
            if type(json) != str and type(json) != bytes:
                try:
                    json = jsonlib.dumps(json)
                except TypeError as e:
                    raise InternalTypeError("failed to serialize 'json' input") from e

        # multipart request preparation
        me: Optional[MultipartEncoder] = None
        if multipart:
            data: Dict[str, Any] = {}
            # if json will be sent as multipart clean it up
            if json is not None:
                if not isinstance(json, dict):
                    raise InternalTypeError(f"Received 'json' parameter with a multipart request but was of type "
                                            f"{type(json)} not dict.")
                for field_name, field_value in json.items():
                    # validate key type
                    if not isinstance(field_name, str):
                        raise InternalTypeError(f"both 'files' and 'json' were supplied but 'json' keys type is "
                                                f"{type(field_name)} not str")
                    # serialize the values if needed
                    if isinstance(field_value, dict) or isinstance(field_value, list):
                        try:
                            field_value = jsonlib.dumps(field_value)
                        except TypeError as e:
                            raise InternalTypeError(f"failed to serialize 'json' input for key '{field_name}'") \
                                from e
                    # convert strings to BytesIO
                    if isinstance(field_value, str):
                        field_value = BytesIO(bytes(field_value, encoding='utf-8'))

                    # check that our final value is file-like
                    try:
                        field_value.seek(0)
                    except AttributeError as e:
                        raise InternalTypeError(f"Received 'data' dict but could not convert field '{field_name}' "
                                                f"of type '{type(json[field_name])}' to file-like object") from e
                    # set value in our newly-built data
                    data[field_name] = field_value

            if files is not None:
                # if list, must be of tuples like ("fname", data, [encoding]) -- add to data in dict format
                if isinstance(files, list):
                    for entry in files:
                        if not (isinstance(entry, tuple) and len(entry) >= 2 and isinstance(entry[0], str)):
                            raise InternalTypeError(f"received list for files argument but did not contain tuples in "
                                                    f"the correct format, entry was of type {type(entry)}: {entry}")
                        data[entry[0]] = entry
                # if dict, ensure in tuple format like ("fname", data, [encoding]) or reformat if not
                elif isinstance(files, dict):
                    for fname in files.keys():
                        file_obj = files[fname]
                        if isinstance(file_obj, tuple):
                            data[fname] = file_obj
                        elif hasattr(file_obj, 'read'):
                            data[fname] = (fname, file_obj)
                        else:
                            raise InternalTypeError(f"files['{fname}'] is of type {type(file_obj)}, not a tuple or "
                                                    f"file-like")
                else:
                    raise InternalTypeError(
                        f"received 'files' argument but was of type '{type(files)}; not list or dict")

        url = construct_url(self.api_base_url, endpoint)

        # REQUEST SENDING AND RESPONSE PARSING
        resp = None
        success = False
        error = None
        attempts = 0
        while not success and attempts <= retries:
            if attempts > 0:
                wait_time = self.BACKOFF_CONSTANT * (self.BACKOFF_EXPONENT_BASE ** attempts)
                sleep(wait_time)
            try:
                if multipart:
                    # create multipart encoder
                    #  note: it will re-seek to the beginning of all files so retries are safe
                    me = MultipartEncoder(fields=data)
                    headers['Content-Type'] = me.content_type
                    resp = self.session.request(method, url, params=params, data=me, headers=headers,
                                                timeout=self.timeout)
                else:
                    resp = self.session.request(method, url, params=params, headers=headers, data=json,
                                                timeout=self.timeout)
                if validate_response_status:
                    validation.validate_response_status(resp, expected_status_code=validation_response_code)
                success = True
            except (ResponseServerError, ResponseRedirectError, RequestException) as e:
                error = e
                logger.debug(e)

            attempts += 1

        if not success:
            if error is None:
                # should never happen but just in case
                raise ArthurInternalError(f"failed to send request but error is None")
            raise error

        if resp is None:
            raise InternalTypeError("Response object was None")
        elif return_raw_response:
            return resp
        else:
            return _parse_response(resp)

    def get(self,
            endpoint: str,
            headers: Optional[Dict[str, str]] = None,
            params: Optional[Union[Dict, bytes]] = None,
            return_raw_response: bool = False,
            retries: int = 0,
            validate_response_status: bool = True,
            validation_response_code: Optional[int] = None) -> Union[Dict, List, bytes, BytesIO, requests.Response]:
        """
        Send an HTTP GET request
        :param endpoint: the specific endpoint to append to the client URL
        :param headers: headers to use for this request in addition to the client default headers
        :param params: query parameters to add to the request
        :param return_raw_response: if true, return the requests.Response object received; otherwise attempt to parse
                                    the response
        :param retries: number of times to retry the request on failure. uses exponential backoff
        :param validate_response_status: if True, raise an ArthurException if the status code is not 2XX or does not
            match `validation_response_code`
        :param validation_response_code: expected status code of the response to validate. if None, allow any 2XX
        :return: if return_raw_response is true, return the requests.Response object received; otherwise attempt to
        parse the response
        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        return self.send(endpoint, method='GET', headers=headers, params=params,
                         return_raw_response=return_raw_response,
                         retries=retries, validate_response_status=validate_response_status,
                         validation_response_code=validation_response_code)

    def post(self,
             endpoint: str,
             json: Optional[Union[Dict, List, str, bytes]] = None,
             files: Optional[Union[Dict[str, BinaryIO], List[Tuple], Dict[str, Tuple]]] = None,
             headers: Optional[Dict[str, str]] = None,
             params: Optional[Union[Dict, bytes]] = None,
             return_raw_response: bool = False,
             retries: int = 0,
             validate_response_status: bool = True,
             validation_response_code: Optional[int] = None) -> Union[Dict, List, bytes, BytesIO, requests.Response]:
        """
        Send an HTTP POST request
        :param endpoint: the specific endpoint to append to the client URL
        :param headers: headers to use for this request in addition to the client default headers
        :param json: data to send as JSON, either a string/bytes to send directly or a dictionary/list to serialize. if
                     `files` is also supplied, this should be a map from name to content, to be sent along with the
                     `files` as a multipart request
        :param files: a map from file names to file-like objects, to be sent as multipart/form-data
        :param params: query parameters to add to the request
        :param return_raw_response: if true, return the requests.Response object received; otherwise attempt to parse
                                    the response
        :param retries: number of times to retry the request on failure. uses exponential backoff
        :param validate_response_status: if True, raise an ArthurException if the status code is not 2XX or does not
            match `validation_response_code`
        :param validation_response_code: expected status code of the response to validate. if None, don't validate
        :return: if return_raw_response is true, return the requests.Response object received; otherwise attempt to
        parse the response
        """
        return self.send(endpoint, method='POST', headers=headers, json=json, files=files, params=params,
                         return_raw_response=return_raw_response, retries=retries,
                         validate_response_status=validate_response_status,
                         validation_response_code=validation_response_code)

    def put(self,
            endpoint: str,
            json: Optional[Union[Dict, List, str, bytes]] = None,
            files: Optional[Union[Dict[str, BinaryIO], List[Tuple], Dict[str, Tuple]]] = None,
            headers: Optional[Dict[str, str]] = None,
            params: Optional[Union[Dict, bytes]] = None,
            return_raw_response: bool = False,
            retries: int = 0,
            validate_response_status: bool = True,
            validation_response_code: Optional[int] = None) -> Union[Dict, List, bytes, BytesIO, requests.Response]:
        """
        Send an HTTP PUT request
        :param endpoint: the specific endpoint to append to the client URL
        :param headers: headers to use for this request in addition to the client default headers
        :param json: data to send as JSON, either a string/bytes to send directly or a dictionary/list to serialize. if
                     `files` is also supplied, this should be a map from name to content, to be sent along with the
                     `files` as a multipart request
        :param files: a map from file names to file-like objects, to be sent as multipart/form-data
        :param params: query parameters to add to the request
        :param return_raw_response: if true, return the requests.Response object received; otherwise attempt to parse
                                    the response
        :param retries: number of times to retry the request on failure. uses exponential backoff
        :param validate_response_status: if True, raise an ArthurException if the status code is not 2XX or does not
            match `validation_response_code`
        :param validation_response_code: expected status code of the response to validate. if None, don't validate
        :return: if return_raw_response is true, return the requests.Response object received; otherwise attempt to
        parse the response
        """
        return self.send(endpoint, method='PUT', headers=headers, json=json, files=files, params=params,
                         return_raw_response=return_raw_response, retries=retries,
                         validate_response_status=validate_response_status,
                         validation_response_code=validation_response_code)

    def patch(self,
              endpoint: str,
              json: Optional[Union[Dict, List, str, bytes]] = None,
              files: Optional[Union[Dict[str, BinaryIO], List[Tuple], Dict[str, Tuple]]] = None,
              headers: Optional[Dict[str, str]] = None,
              params: Optional[Union[Dict, bytes]] = None,
              return_raw_response: bool = False,
              retries: int = 0,
              validate_response_status: bool = True,
              validation_response_code: Optional[int] = None) -> Union[Dict, List, bytes, BytesIO, requests.Response]:
        """
        Send an HTTP POST request
        :param endpoint: the specific endpoint to append to the client URL
        :param headers: headers to use for this request in addition to the client default headers
        :param json: data to send as JSON, either a string/bytes to send directly or a dictionary/list to serialize. if
                     `files` is also supplied, this should be a map from name to content, to be sent along with the
                     `files` as a multipart request
        :param files: a map from file names to file-like objects, to be sent as multipart/form-data
        :param params: query parameters to add to the request
        :param return_raw_response: if true, return the requests.Response object received; otherwise attempt to parse
                                    the response
        :param retries: number of times to retry the request on failure. uses exponential backoff
        :param validate_response_status: if True, raise an ArthurException if the status code is not 2XX or does not
            match `validation_response_code`
        :param validation_response_code: expected status code of the response to validate. if None, don't validate
        :return: if return_raw_response is true, return the requests.Response object received; otherwise attempt to
        parse the response
        """
        return self.send(endpoint, method='PATCH', headers=headers, json=json, files=files, params=params,
                         return_raw_response=return_raw_response, retries=retries,
                         validate_response_status=validate_response_status,
                         validation_response_code=validation_response_code)

    def delete(self,
               endpoint: str,
               headers: Optional[Dict[str, str]] = None,
               return_raw_response: bool = False,
               retries: int = 0,
               validate_response_status: bool = True,
               validation_response_code: Optional[int] = None) -> Union[Dict, List, bytes, BytesIO, requests.Response]:
        """
        Send an HTTP DELETE request
        :param endpoint: the specific endpoint to append to the client URL
        :param headers: headers to use for this request in addition to the client default headers
        :param return_raw_response: if true, return the requests.Response object received; otherwise attempt to parse
                                    the response
        :param retries: number of times to retry the request on failure. uses exponential backoff
        :param validate_response_status: if True, raise an ArthurException if the status code is not 2XX or does not
            match `validation_response_code`
        :param validation_response_code: expected status code of the response to validate. if None, don't validate
        :return: if return_raw_response is true, return the requests.Response object received; otherwise attempt to
        parse the response
        """
        return self.send(endpoint, method='DELETE', headers=headers, return_raw_response=return_raw_response,
                         retries=retries, validate_response_status=validate_response_status,
                         validation_response_code=validation_response_code)



import unittest

import logging

import pytest
import responses
import json

# import client library
from arthurai import ArthurAI
from tests.fixtures.mocks import BASE_URL


class BaseTest(unittest.TestCase):
    _base_url = BASE_URL
    _content_type_json = 'application/json'

    @classmethod
    def setUpClass(cls):
        logging.info("Base Set up")

        cls.access_key = "FAKE_ACCESS_KEY"

        config = {
            'access_key': cls.access_key,
            'url': BaseTest._base_url,
            'offline': True
        }
        cls.arthur = ArthurAI(config)

    @classmethod
    def tearDownClass(cls):
        logging.info("Base Tear down")
        pass

    def mockPost(self, append_url, dict_to_return, request_type=responses.POST, status=200):
        """
        Adds a mock 200 response

        :param append_url: the url to the base url
        :param dict_to_return: the dictionary to return as json
        :return: None
        """

        responses.add(request_type, BaseTest._base_url + append_url,
                      json.dumps(dict_to_return), status=status,
                      content_type=self._content_type_json)

    def mockPost500(self, append_url, request_type=responses.POST):
        responses.add(request_type, BaseTest._base_url + append_url,
                      json.dumps({"Message": "There has been a server side exception"}), status=500,
                      content_type=self._content_type_json)

    def mockGet(self, append_url, response_body, request_type=responses.GET, status=200, content_type=_content_type_json):
        """
        Adds a mock 200 response

        :param append_url: the url to the base url
        :param response_content: the response_content to return from the mock API request
        :return: None
        """

        responses.add(request_type,
                      BaseTest._base_url + append_url,
                      body=response_body,
                      status=status,
                      content_type=content_type)




import logging


DEFAULT_LOG_LEVEL = logging.INFO


def initialize_logging():
    """Creates a parent 'arthurai' logger with a console output and INFO level.
    """
    arthur_logger = logging.getLogger("arthurai")
    arthur_logger.setLevel(DEFAULT_LOG_LEVEL)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)

    arthur_logger.handlers = []
    arthur_logger.addHandler(ch)


def _set_arthur_log_level(level: int):
    logging.getLogger("arthurai").setLevel(level)


def enable_debug_logging():
    """Enables debug logging for the arthurai package.
    """
    _set_arthur_log_level(logging.DEBUG)


def disable_debug_logging():
    """Disables debug logging for the arthurai package.
    """
    _set_arthur_log_level(DEFAULT_LOG_LEVEL)



import json
import logging
import os
from pathlib import Path
import re
from typing import Dict, Optional, Union, Any, List, Sequence, Iterable

import numpy as np
import pandas as pd

import arthurai.util as arthur_util
from arthurai.common.constants import ValueType
from arthurai.common.exceptions import InternalValueError, InternalTypeError, UserTypeError

logger = logging.getLogger(__name__)


def retrieve_parquet_files(directory_path: str) -> List[Path]:
    """Checks whether a given directory and its subdirectories contain parquet files,
    if so this will return a list of the files

    :param directory_path:    local path to check files types

    :return: List of paths for parquet files that are found
    """
    parquet_files = []
    for path, subdir, files in os.walk(directory_path):
        for file in files:
            if file.endswith('.parquet'):
                parquet_files.append(Path(os.path.join(path, file)))
    return parquet_files


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    @staticmethod
    def convert_value(obj):
        """Converts the given object from a numpy data type to a python data type, if the object is already a
        python data type it is returned

        :param obj: object to convert
        :return: python data type version of the object
        """
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64,
                            np.long, np.longlong)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        elif isinstance(obj, pd.Timestamp):
            return obj.isoformat()
        elif isinstance(obj, float) and np.isnan(obj):
            return None
        return obj


def standardize_pd_obj(data: Union[pd.DataFrame, pd.Series], dropna: bool, replacedatetime: bool,
                       attributes: Optional[Dict[str, Union[str, ValueType]]] = None) -> Union[pd.DataFrame, pd.Series]:
    """Standardize pandas object for nans and datetimes.
    
    Standardization includes casting correct type for int columns that are float due to nans and 
    for converting datetime objects into isoformatted strings.

    :param data: the pandas data to standardize
    :param dropna: if True, drop nans from numeric date columns
    :param replacedatetime: if True, replace timestamps with isoformatted strings
    :param attributes: if used for sending inferences, will handle column type conversions for columns with any nulls

    :return: the standardized pandas data
    :raise: TypeError: timestamp is not of type `datetime.datetime`
    :raise: ValueError: timestamp is not timezone aware and no location data is provided to remedy
    """
    
    def standardize_pd_series(series: pd.Series, datatype: Optional[str]) -> pd.Series:
        series = series.replace([np.inf, -np.inf], np.nan)
        nans_present = series.isnull().values.any()
        if dropna:
            series = series.dropna()
            if len(series) == 0:
                return series

        # handle case where int column has nans and therefore is incorrectly seen as float column
        if re.search("float*", series.dtype.name, flags=re.I) and nans_present and datatype == ValueType.Integer:
            valid_series = series.dropna()
            if len(valid_series) > 0 and np.array_equal(valid_series, valid_series.astype(int)):
                return series.astype(pd.Int64Dtype())

        # check to make sure datetimes are timezone aware
        # TODO: remove is_date and first conditional below when deprecating string support
        #  @ https://arthurai.atlassian.net/jira/software/projects/CR/boards/11?selectedIssue=CR-10
        elif re.search("(datetime|timestamp).*", series.dtype.name, flags=re.I) or \
                arthur_util.is_string_date(series.values[0]):
            if arthur_util.is_string_date(series.values[0]):
                logger.warning(
                    f"DEPRECATION WARNING: Your feature <{series.name}> providing timestamps starting with "
                    f"<{series.values[0]}> is of type `str` and is assumed to be in UTC format. `timezone` information "
                    f"will be ignored. `str` types will be deprecated. Please use timeaware `datetime.datetime` "
                    f"timestamps and use arthurai.util.format_timestamp if needed, e.g. "
                    f"`format_timestamp(datetime.strptime('2020-01-01 01:01:01', '%Y-%m-%d %H:%M:%S'), 'US/Eastern')`."
                )
            formatted_series = series.apply(arthur_util.format_timestamp)
            if replacedatetime:
                return formatted_series

        return series

    if isinstance(data, pd.Series):
        datatype = None
        if attributes and data.name in attributes:
            datatype = attributes[data.name]
        return standardize_pd_series(data, datatype)

    elif isinstance(data, pd.DataFrame):
        if dropna:
            raise InternalValueError(f"Cannot use dropna={dropna} with data argument as pd.DataFrame.")
        df = data.copy()
        for column in df.columns:
            datatype = None
            if attributes and column in attributes:
                datatype = attributes[column]
            df[column] = standardize_pd_series(df[column], datatype)
        return df

    else:
        raise InternalTypeError("Cannot standardize object that is not pd.DataFrame or pd.Series.")


def dataframe_like_to_list_of_dicts(data: Union[List[Dict[str, Any]], Dict[str, List[Any]], pd.DataFrame]):
    """
    Standardize data in a List of Dicts format (e.g. `[{'a': 1, 'b': 2}, {'a': 3, 'b': 4}]`). Input can be formatted as
    a List of Dicts, Dict of Lists (e.g. `{'a': [1, 3], 'b': [2, 4]}`), or a Pandas DataFrame. May return the same
    object as input if it already matches the correct format.
    :param data: the input data to format
    :return: the data restructured as a Dict of Lists
    :raise UserTypeError: if the data is in an unexpected format
    """
    if len(data) == 0:
        return []

    if isinstance(data, list) and isinstance(data[0], dict):
        return data
    elif isinstance(data, dict):
        data = pd.DataFrame(data)

    if isinstance(data, pd.DataFrame):
        return data.to_dict(orient='records')
    else:
        raise UserTypeError(f"Invalid input type {type(data)}, should be list of dicts, dict of lists, or DataFrame")


def update_column_in_list_of_dicts(data: List[Dict[str, Any]], target_column: str, column_values: Sequence[Any]) -> \
        None:
    """
    Adds column_values to target_column in a list of dicts **in place**. If values are present for target_column, they are
    overwritten.
    :param data: the List of Dict data to modify
    :param target_column: the name of the column to write values into
    :param column_values: the values to write
    :return: None (in place)
    :raises UserValueError: if the lengths don't match or aren't retrievable
    """
    try:
        if len(column_values) != len(data):
            raise ValueError(f"The dataset has {len(data)} elements but column {target_column} has "
                             f"{len(column_values)} elements. Cannot add values to data.")
    except TypeError:
        raise UserTypeError(f"Cannot compare lengths of data and values for column {target_column}, are you sure the "
                            "values have a definite length?")
    for i in range(len(data)):
        data[i][target_column] = column_values[i]


def intersection_is_non_empty(iter1: Iterable[Any], iter2: Iterable[Any]):
    """
    Returns True if the two iterables share at least one element
    :param iter1:
    :param iter2:
    :return:
    """
    for val in iter1:
        if val in iter2:
            return True
    return False



from typing import Optional, List
import os
from distutils.util import strtobool
from platform import system
from http import HTTPStatus
import logging

from arthurai.client.helper import get_current_user_org
from arthurai.client.http.arthur import FALSE_OFFLINE_DEFAULT
from arthurai.client.http.requests import HTTPClient
from arthurai.common.constants import InputType, OutputType, TextDelimiter, Role, DEFAULT_SERVICE_ACCOUNT, API_PREFIX
from arthurai.common.exceptions import MissingParameterError, UserTypeError, UserValueError, arthur_excepted
from arthurai.core.models import ArthurModel
from arthurai.client.validation import validate_response_status
from arthurai.version import __version__ as arthur_version
from arthurai.client.helper import user_login

logger = logging.getLogger(__name__)


def new_requests_client(access_key: str = None, url: str = None, verify_ssl: bool = True,
                        offline: bool = False) -> HTTPClient:
    if access_key is None:
        access_key = os.getenv('ARTHUR_API_KEY')
    if url is None:
        url = os.getenv('ARTHUR_ENDPOINT_URL')

    if url is None or access_key is None:
        raise MissingParameterError("Please set api key and url either via environment variables "
                                    "(`ARTHUR_API_KEY` and `ARTHUR_ENDPOINT_URL`) or by passing parameters "
                                    "`access_key` and `url`.")

    if offline:
        org_id = "unknown"
    else:
        org_id = get_current_user_org(url, API_PREFIX, access_key, verify_ssl=verify_ssl)
    user_agent = f"arthur-sdk/{arthur_version} (system={system()}, org={org_id})"
    headers = {
        'Accept': "application/json",
        'Authorization': access_key,
        'User-Agent': user_agent
    }
    return HTTPClient(base_url=url, path_prefix=API_PREFIX, default_headers=headers, verify_ssl=verify_ssl)



class ArthurAI(object):
    """A client that interacts with Arthur's servers."""

    def __init__(self, config=None, verify_ssl=None, url=None, access_key=None, offline=FALSE_OFFLINE_DEFAULT,
                 login=None, password=None):

        verify_ssl_env_var = os.getenv('VERIFY_SSL')
        if verify_ssl is not None:
            if not isinstance(verify_ssl, bool):
                raise UserTypeError("verify_ssl must be of type bool")
            self.verify_ssl = verify_ssl
        elif verify_ssl_env_var:
            try:
                self.verify_ssl = bool(strtobool(verify_ssl_env_var))
            except ValueError:
                raise UserValueError("Environment Variable VERIFY_SSL must be one of True, true, T, t, False, false, F, f, 1, 0")
        else:
            self.verify_ssl = True

        if config:
            self.client = new_requests_client(verify_ssl=self.verify_ssl, **config)

        elif login and password:
            # Get temporary auth token from username and password
            auth_token, _ = user_login(api_http_host=url, api_prefix=API_PREFIX, login=login, password=password,
                                       verify_ssl=self.verify_ssl)
            # Define a client with the temporary auth token
            self.client = new_requests_client(access_key=auth_token, url=url, verify_ssl=self.verify_ssl,
                                              offline=offline)
            # Check to see if the default service account exitst, if not create it
            service_account = self.get_service_accounts(description=DEFAULT_SERVICE_ACCOUNT)
            if not service_account:
                service_account = [self.create_service_account(description=DEFAULT_SERVICE_ACCOUNT,
                                                                role=Role.ModelOwner)]

            # Update the client access key with the permanent key from the database
            self.client.access_key = service_account[0]['token']
        else:
            self.client = new_requests_client(access_key=access_key, url=url, verify_ssl=self.verify_ssl,
                                              offline=offline)

    def model(self,
              partner_model_id: str,
              input_type: InputType,
              output_type: Optional[OutputType] = None,
              model_type: Optional[OutputType] = None,
              display_name: Optional[str] = None,
              description: Optional[str] = None,
              tags: Optional[List[str]] = None,
              classifier_threshold: Optional[float] = None,
              is_batch: bool = False,
              text_delimiter: Optional[TextDelimiter] = None,
              expected_throughput_gb_per_day: Optional[int] = None,
              pixel_height: Optional[int] = None,
              pixel_width: Optional[int] = None) -> ArthurModel:
        """Create a new multistage model.

        :param partner_model_id: The string external id of the model
        :param input_type: the :py:class:`.InputType`
        :param output_type: the :py:class:`.OutputType`
        :param model_type: .. deprecated:: version 2.0.0 Use `output_type` instead.
        :param display_name: Optional name to display on dashboard, will default to the external id
        :param description: Optional description for the model
        :param tags: A list of string tags to associate with the model
        :param classifier_threshold: For binary classification models this is the threshold to determine a positive class, defaults to 0.5
        :param is_batch: boolean value which signifies whether the model sends inferences by batch or streaming
        :param text_delimiter: TextDelimiter used in NLP models to split documents into tokens for explanations
        :param expected_throughput_gb_per_day: Expected amount of throughput. Used to provision resources
        :param pixel_height: Image height in pixels. Needed for CV models which require images to be one size
        :param pixel_width: Image width in pixels. Needed for CV models which require images to be one size

        :return: An :py:class:`~arthurai.client.apiv3.model.ArthurModel`
        """
        if output_type is None and model_type is None:
            raise MissingParameterError("Either 'output_type' or 'model_type' parameter must be specified")
        output_type = output_type if output_type is not None else model_type

        return ArthurModel(
            client=self.client,
            partner_model_id=partner_model_id,
            input_type=input_type,
            output_type=output_type,
            display_name=display_name,
            description=description,
            tags=tags,
            classifier_threshold=classifier_threshold,
            is_batch=is_batch,
            text_delimiter=text_delimiter,
            expected_throughput_gb_per_day=expected_throughput_gb_per_day,
            pixel_height=pixel_height,
            pixel_width=pixel_width,
        )

    @arthur_excepted("failed to retrieve model")
    def get_model(self, identifier: str, id_type: str = 'id') -> ArthurModel:
        """Retrieve an existing model by id

        :param: identifier: Id to get the model by
        :param: id_type: Type of id the identifier is, possible options are ['id', 'partner_model_id']

        :raise: ArthurUserError: failed due to user error
        :raise: ArthurInternalError: failed due to an internal error
        """
        if id_type in ['partner_model_id', 'id']:
            endpoint = f'/models/{identifier}?id_type={id_type}&expand=attributes'
        else:
            raise UserValueError(f"Invalid id_type: {id_type}, must be one of the following ['id', 'partner_model_id]")

        resp = self.client.get(endpoint, return_raw_response=True, validation_response_code=HTTPStatus.OK)

        # accuracy data is only used by the UI not needed in the SDK
        model_data = resp.json()
        if "accuracy_enabled" in model_data:
            del model_data["accuracy_enabled"]

        model = ArthurModel.from_dict(model_data)
        model._client = self.client
        return model

    @arthur_excepted("failed to get service accounts")
    def get_service_accounts(self, description: Optional[str] = None):
        """
        Retrieve service accounts
        :param description: an optional description that returned account should match
        :return: parsed response containing a list of service accounts, in the following format:
            ```python
            [
                {
                    "id": "<service account id>",
                    "organization_id": "<organization id>",
                    "description": "<service account description>",
                    "token": "<service account api key>"
                }
            ]
            ```
        """
        endpoint = "/service_accounts"
        params = None if description is None else {'description': description}
        resp = self.client.get(endpoint=endpoint, params=params, return_raw_response=True)
        validate_response_status(resp, expected_status_code=HTTPStatus.OK)

        return resp.json()['data']

    @arthur_excepted("failed to create service account")
    def create_service_account(self, description: str, role: Role):
        """
        Create a new service account
        :param description: the description for the new service account
        :param role: the Role of the new service account
        :return: parsed response in the following format:
            ```python
            {
                "id": "<service account id>",
                "organization_id": "<organization id>",
                "description": "<service account description>",
                "token": "<service account api key>"
            }
            ```
        """
        endpoint = "/service_accounts"
        body = {
            "description": description,
            "role": role
        }
        resp = self.client.post(endpoint=endpoint, json=body, return_raw_response=True)
        validate_response_status(resp, expected_status_code=HTTPStatus.CREATED)

        return resp.json()



import os
import re
from http import HTTPStatus

import pytest
import responses

from tests import MockResponse

from arthurai import ArthurAI
from arthurai.common.constants import InputType, OutputType
from arthurai.client.validation import validate_multistatus_response_and_get_failures, validate_response_status
from arthurai.common.exceptions import ResponseServerError, ResponseClientError, ResponseRedirectError, \
    InternalValueError, UserValueError, UserTypeError
from tests.fixtures.mocks import ACCESS_KEY, BASE_URL, USER_ACCESS_KEY
from tests.helpers import mock_get, mock_post
from unittest.mock import patch


class TestClient:

    def _mock_get_user(self):
        mock_get(f"/api/v3/users/me", status=HTTPStatus.OK,
                 response_body={"organization_id": "65624ee6-9a73-4fe3-a7ff-8612f7d11c21", "role": "Model Owner"})

    def _mock_get_user_unauthorized(self):
        mock_get(f"/api/v3/users/me", status=HTTPStatus.UNAUTHORIZED,
                 response_body={"bad": "user"})

    def _mock_user_login(self):
        mock_post(f"/api/v3/login", status=HTTPStatus.OK, response_body={},
                  headers={"Set-Cookie": 'Authorization=' + USER_ACCESS_KEY})

    @pytest.mark.usefixtures('mock_cred_env_vars')
    def test_client_init_env_vars(self):
        client = ArthurAI(offline=True)
        assert client.client.api_base_url == 'http://mock/api/v3'
        assert client.client.session.headers.get('Authorization') == 'access_key'
        assert client.client.session.verify is False

        client = ArthurAI(url="this is a url", access_key="this is an access key", offline=True)
        assert client.client.api_base_url == "https://this is a url/api/v3"
        assert client.client.session.headers.get('Authorization') == "this is an access key"
        assert client.client.session.verify is False

        client = ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True)
        assert client.client.api_base_url == "http://this is a url/api/v3"
        assert client.client.session.headers.get('Authorization') == "this is an access key"
        assert client.client.session.verify is False

        with pytest.raises(ValueError):
            ArthurAI(url="http://this is a url/apath", access_key="this is an access key", offline=True)

        with pytest.raises(ValueError):
            ArthurAI(url="http://this is a url?food=good", access_key="this is an access key", offline=True)

        with pytest.raises(ValueError):
            ArthurAI(url="fatfinger://this is a url", access_key="this is an access key", offline=True)

        with pytest.raises(ValueError):
            ArthurAI(url="https:", access_key="this is an access key", offline=True)

    @pytest.mark.usefixtures('mock_cred_env_vars_ssl_true')
    def test_client_init_env_vars_ssl_true(self):
        client = ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True)
        assert client.client.session.verify is True

    @pytest.mark.usefixtures('mock_cred_env_vars_ssl_garbage')
    def test_client_init_env_vars_ssl_garbage(self):
        with pytest.raises(UserValueError):
            ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True)

    @responses.activate
    def test_client_init(self):
        self._mock_get_user()

        ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        assert len(responses.calls) == 1
        expectedResponse = {'organization_id': '65624ee6-9a73-4fe3-a7ff-8612f7d11c21', 'role': 'Model Owner'}
        assert expectedResponse == responses.calls[0].response.json()

    @responses.activate
    def test_client_init_config_param(self):
        self._mock_get_user()

        ArthurAI(config={'access_key': ACCESS_KEY, 'url': BASE_URL})
        assert len(responses.calls) == 1
        expectedResponse = {'organization_id': '65624ee6-9a73-4fe3-a7ff-8612f7d11c21', 'role': 'Model Owner'}
        assert expectedResponse == responses.calls[0].response.json()

    @responses.activate
    def test_client_init_bad_host(self):
        with pytest.raises(UserValueError, match=".*please ensure the URL is correct.*"):
            ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        assert len(responses.calls) == 1

    @responses.activate
    def test_client_init_bad_access_key(self):
        self._mock_get_user_unauthorized()
        with pytest.raises(UserValueError, match=".*please ensure your access key is correct.*"):
            ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        assert len(responses.calls) == 1
    
    @responses.activate
    def test_old_client_login_with_username_password(self):
        self._mock_user_login()
        self._mock_get_user()

        with patch("arthurai.client.client.ArthurAI.get_service_accounts") as get_sa_mock:
            get_sa_mock.return_value = [{"id": "", "organization_id": "", "description": "",
                                            "token": ACCESS_KEY}]
            arthur = ArthurAI(login="arthur", password="password", url=BASE_URL)
            assert ACCESS_KEY == arthur.client.access_key

    @responses.activate
    def test_new_client_login_with_username_password(self):
        self._mock_user_login()
        self._mock_get_user()

        with patch("arthurai.client.client.ArthurAI.get_service_accounts") as get_sa_mock:
            get_sa_mock.return_value = None
            with patch("arthurai.client.client.ArthurAI.create_service_account") as get_create_mock:
                get_create_mock.return_value = {"id": "", "organization_id": "",
                                                    "description": "", "token": ACCESS_KEY}
                arthur = ArthurAI(login="arthur", password="password", url=BASE_URL)
                assert ACCESS_KEY == arthur.client.access_key
                assert arthur.verify_ssl is True
                assert arthur.client.session.verify is True

                arthur = ArthurAI(login="arthur", password="password", url=BASE_URL, verify_ssl=False)
                assert arthur.client.session.verify is False

    @responses.activate
    def test_client_init_offline(self):
        ArthurAI(access_key=ACCESS_KEY, url=BASE_URL, offline=True)
        assert len(responses.calls) == 0

    @responses.activate
    def test_client_init_offlines(self):
        mock_get(f"/api/v3/models/1234", status=200,
                 response_body={"organization_id": "65624ee6-9a73-4fe3-a7ff-8612f7d11c21", "role": "Model Owner"})
        ArthurAI(access_key=ACCESS_KEY, url=BASE_URL, offline=True)
        assert len(responses.calls) == 0

    @responses.activate
    def test_client_sets_model_id_and_user_header(self):
        model_data = {
            "id": "1234",
            "partner_model_id": "",
            "input_type": InputType.Tabular,
            "output_type": OutputType.Regression,
            "display_name": "",
            "description": "",
            "attributes": [],
        }
        mock_get("/api/v3/models/1234", status=200, response_body=model_data)

        self._mock_get_user()

        client = ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        client.get_model('1234')

        assert len(responses.calls) == 2
        if "models" in responses.calls[1].request.url:
            header = responses.calls[1].request.headers
        else:
            header = responses.calls[0].request.headers
        assert 'User-Agent' in header
        print(header['User-Agent'])
        assert re.findall(r"^arthur-sdk\/\d.\d+.\d+ \(system=\w*, org=65624ee6-9a73-4fe3-a7ff-8612f7d11c21\)$",
                          header['User-Agent'])
        assert os.getenv("ARTHUR_LAST_MODEL_ID") == "1234"


class TestGetMultiStatusResponseFailures:
    # fixtures
    SUCCESS_RESULTS = [{"message": "ok", "status": HTTPStatus.OK}, {"message": "created", "status": HTTPStatus.CREATED}]
    USER_FAIL_RESULTS = [{"message": "not found", "status": HTTPStatus.NOT_FOUND}]
    INTERNAL_FAIL_RESULTS = [{"message": "failure", "status": HTTPStatus.INTERNAL_SERVER_ERROR},
                             {"message": "choose", "status": HTTPStatus.MULTIPLE_CHOICES}]
    SUCCESS_BODY = {"counts": {"success": len(SUCCESS_RESULTS), "failure": 0, "total": len(SUCCESS_RESULTS)},
                    "results": SUCCESS_RESULTS}
    FAILURES_BODY = {"counts": {"success": len(SUCCESS_RESULTS),
                                "failure": len(INTERNAL_FAIL_RESULTS) + len(USER_FAIL_RESULTS),
                                "total": len(SUCCESS_RESULTS) + len(INTERNAL_FAIL_RESULTS) + len(USER_FAIL_RESULTS)},
                     "results": SUCCESS_RESULTS + USER_FAIL_RESULTS + INTERNAL_FAIL_RESULTS}
    MISMATCHED_COUNTS_BODY = {"counts": {"success": len(SUCCESS_RESULTS), "failure": 0,
                                         "total": len(SUCCESS_RESULTS)},
                              "results": SUCCESS_RESULTS + USER_FAIL_RESULTS + INTERNAL_FAIL_RESULTS}

    def test_success(self):
        response = MockResponse(self.SUCCESS_BODY, HTTPStatus.MULTI_STATUS)
        actual_user_failures, actual_internal_failures = validate_multistatus_response_and_get_failures(response)
        assert actual_user_failures == []
        assert actual_internal_failures == []

    def test_failures(self):
        response = MockResponse(self.FAILURES_BODY, HTTPStatus.MULTI_STATUS)
        actual_user_failures, actual_internal_failures = validate_multistatus_response_and_get_failures(response)
        assert actual_user_failures == self.USER_FAIL_RESULTS
        assert actual_internal_failures == self.INTERNAL_FAIL_RESULTS

    def test_bad_status(self):
        response = MockResponse(self.SUCCESS_BODY, HTTPStatus.OK)
        with pytest.raises(ValueError):
            validate_multistatus_response_and_get_failures(response)

    def test_mismatched_counts(self):
        response = MockResponse(self.MISMATCHED_COUNTS_BODY, HTTPStatus.MULTI_STATUS)
        with pytest.raises(ValueError):
            validate_multistatus_response_and_get_failures(response)


# TODO: add testing around allowed exceptions and what gets raised first (e.g. expected 207 but 404 shouldn't raise
#  InternalValueError)
class TestValidateResponse:
    # fixtures
    JSON_BODY = {"status": "somestatus"}
    BYTES_BODY = b"bodybytes"

    def test_server_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.INTERNAL_SERVER_ERROR)
        with pytest.raises(ResponseServerError):
            validate_response_status(response)

    def test_client_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.NOT_FOUND)
        with pytest.raises(ResponseClientError):
            validate_response_status(response)

    def test_expected_404_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.NOT_FOUND)
        validate_response_status(response, expected_status_code=HTTPStatus.NOT_FOUND)

    def test_redirect_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.MOVED_PERMANENTLY)
        with pytest.raises(ResponseRedirectError):
            validate_response_status(response)

    def test_redirect_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.MOVED_PERMANENTLY)
        validate_response_status(response, allow_redirects=True)

    def test_ok_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.OK)
        validate_response_status(response)

    def test_unexpected_ok_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.OK)
        with pytest.raises(InternalValueError):
            validate_response_status(response, expected_status_code=HTTPStatus.CREATED)

    def test_server_error(self):
        code = HTTPStatus.INTERNAL_SERVER_ERROR
        with pytest.raises(ResponseServerError):
            validate_response_status(code)

    def test_client_error(self):
        code = HTTPStatus.NOT_FOUND
        with pytest.raises(ResponseClientError):
            validate_response_status(code)

    def test_expected_404_no_error(self):
        code = HTTPStatus.NOT_FOUND
        validate_response_status(code, expected_status_code=HTTPStatus.NOT_FOUND)

    def test_redirect_error(self):
        code = HTTPStatus.MOVED_PERMANENTLY
        with pytest.raises(ResponseRedirectError):
            validate_response_status(code)

    def test_redirect_no_error(self):
        code = HTTPStatus.MOVED_PERMANENTLY
        validate_response_status(code, allow_redirects=True)

    def test_ok_no_error(self):
        code = HTTPStatus.OK
        validate_response_status(code)

    def test_unexpected_ok_error(self):
        code = HTTPStatus.OK
        with pytest.raises(InternalValueError):
            validate_response_status(code, expected_status_code=HTTPStatus.CREATED)



from typing import NamedTuple, Tuple, Union, Type

import pytest

from arthurai.client.helper import construct_url
from arthurai.common.exceptions import UserValueError


class ConstructUrlTestCase(NamedTuple):
    parts: Tuple[str, ...]
    validate: bool
    default_https: bool
    expected: Union[str, Type[BaseException]]


CASES = [ConstructUrlTestCase(parts=("https://arthur.ai", "/api/", "/endpoint"), validate=True, default_https=True,
                              expected="https://arthur.ai/api/endpoint"),
         ConstructUrlTestCase(parts=("https://arthur.ai/", "api", "endpoint/value?foo=bar&baz=qux"), validate=True,
                              default_https=True, expected="https://arthur.ai/api/endpoint/value?foo=bar&baz=qux"),
         ConstructUrlTestCase(parts=("arthur.ai", "/api", "/endpoint"), validate=True, default_https=True,
                              expected="https://arthur.ai/api/endpoint"),
         ConstructUrlTestCase(parts=("arthur.ai", "/api", "/endpoint/"), validate=False, default_https=False,
                              expected="arthur.ai/api/endpoint"),
         ConstructUrlTestCase(parts=("arthur.ai", "/api", "/endpoint/"), validate=True, default_https=False,
                              expected=UserValueError),
         ConstructUrlTestCase(parts=("http://arthur.ai/",), validate=True, default_https=True,
                              expected="http://arthur.ai"),
         ConstructUrlTestCase(parts=("foo bar baz",), validate=True, default_https=False, expected=UserValueError)]


@pytest.mark.parametrize("case", CASES)
def test_construct_url(case: ConstructUrlTestCase):
    if isinstance(case.expected, str):
        actual = construct_url(*case.parts, validate=case.validate, default_https=case.default_https)
        assert actual == case.expected
    elif issubclass(case.expected, BaseException):
        with pytest.raises(case.expected):
            construct_url(*case.parts, validate=case.validate, default_https=case.default_https)
    else:
        raise ValueError()



import pandas as pd
import numpy as np
import re
import json

from arthurai.explainability.arthur_explainer import ArthurExplainer
from arthurai.common.constants import InputType, OutputType


def test_arthur_explainer_nlp_regression() -> None:
    def regression_predict(feature_vecs):
        results = []
        for fv in feature_vecs:
            results.append(np.array([0.2, 0.8]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Regression,
                                model_input_type=InputType.NLP,
                                num_predicted_attributes=1,
                                predict_func=regression_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None])

    raw_feat_vecs = [
        ['first test test'],
        ['and a longer second test']
    ]
    exps = explainer.explain_nlp("lime", raw_feature_vectors=raw_feat_vecs, nsamples=100)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        # ensure single class
        assert len(exp) == 1
        # assert one explanation per unique word
        assert len(exp[0]) == len(re.split(explainer.text_delimiter, raw_feat_vecs[i][0]))


def test_arthur_explainer_nlp_classification() -> None:
    def regression_predict(feature_vecs):
        results = []
        for fv in feature_vecs:
            results.append(np.array([0.2, 0.7, 0.1]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Multiclass,
                                model_input_type=InputType.NLP,
                                num_predicted_attributes=3,
                                predict_func=regression_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None])

    raw_feat_vecs = [
        ['first test test'],
        ['and a longer second test']
    ]
    exps = explainer.explain_nlp("lime", raw_feature_vectors=raw_feat_vecs, nsamples=100)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        # ensure single class
        assert len(exp) == 3
        # assert one explanation per unique word
        assert len(exp[0]) == len(re.split(explainer.text_delimiter, raw_feat_vecs[i][0]))


def test_arthur_explainer_image_classification() -> None:
    def load_image(image_path):
        return np.array([
            [[20, 4, 6], [5, 7, 12], [5, 12, 4]],
            [[20, 4, 106], [5, 7, 112], [5, 12, 104]],
            [[20, 4, 206], [5, 7, 212], [5, 12, 204]],
        ])

    def cv_predict(images):
        results = []
        for image in images:
            results.append(np.array([0.2, 0.7, 0.1]))
        return np.array(results)

    sample_data = pd.DataFrame([
        ['this is a test'],
        ['a second test with more words'],
        ['and a third test']
    ])

    explainer = ArthurExplainer(model_type=OutputType.Multiclass,
                                model_input_type=InputType.Image,
                                num_predicted_attributes=3,
                                predict_func=cv_predict,
                                data=sample_data,
                                enable_shap=False,
                                label_mapping=[None],
                                load_image_func=load_image)

    image_paths = [
        'path/to/image',
        'path/to/another/image'
    ]
    exps = explainer.explain_image("lime", image_paths=image_paths)

    # confirm we got 2 explanations
    assert len(exps) == 2

    for i, exp in enumerate(exps):
        """
        Example Output:
            {
                "lime_segment_mask": [
                    [1, 1, 1, 3, 3, 1, 0],
                    [1, 1, 1, 3, 3, 1, 0],
                    [1, 1, 1, 3, 3, 1, 0],
                    [2, 2, 2, 3, 3, 1, 0]
                ],
                "lime_region_mapping": {
                    "0": [[0, 0.23], [1, 0.3], [2, 0.001], [3, -0.5]]
                }
            }
        """
        # confirm we got an appropriate-sized mask
        assert len(exp["lime_segment_mask"]) == len(load_image(image_paths[i]))

        # sanity check on mask values
        # ideally would actually check schema here
        sample_mask_value = exp["lime_segment_mask"][0][0]
        assert sample_mask_value >= 0
        assert next(iter(exp["lime_region_mapping"].values()))

        #ensure serializable
        assert json.dumps(exp)




import json


#  Loads raw JSON
def load_json(path):
    return json.load(open(path))


# Loads a JSON file as a json string
def load_json_string(path):
    json_dict = load_json(path)
    return json.dumps(json_dict)



import responses
import pytest

from arthurai.core.models import ArthurModel
from arthurai.core.bias.bias_metrics import BiasMetrics
from .helpers import mock_post

from .test_request_models.fixtures import model_response_json_strings
from .fixtures.mocks import BASE_URL, ACCESS_KEY, client as mock_client


bias_query_json_result = {
            "query_result": [
            {
                "bias": 1,
                "confusion_matrix": {
                "accuracy_rate": 0.9,
                "balanced_accuracy_rate": 0.8,
                "f1": 0.7,
                "false_negative_rate": 0.5,
                "false_positive_rate": 0,
                "precision": 0.9,
                "true_negative_rate": 1,
                "true_positive_rate": 0.5
                }
            },
            {
                "bias": 2,
                "confusion_matrix": {
                "accuracy_rate": 0.9,
                "balanced_accuracy_rate": 0.8,
                "f1": 0.7,
                "false_negative_rate": 0.4,
                "false_positive_rate": 0,
                "precision": 0.8,
                "true_negative_rate": 1,
                "true_positive_rate": 0.6
                }
            }
            ],
            "sampling_threshold": 1
        }


@pytest.fixture
def binary_model(mock_client):
    binary_model = ArthurModel.from_json(model_response_json_strings[4])
    binary_model._client = mock_client.client
    return binary_model


@pytest.fixture
def multiclass_model(mock_client):
    multiclass_model = ArthurModel.from_json(model_response_json_strings[5])
    multiclass_model._client = mock_client.client
    return multiclass_model


@responses.activate
def test_get_bias_cm_binary(binary_model):
    endpoint = "/api/v3/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasbinary/inferences/query"
    mock_post(endpoint, response_body=bias_query_json_result)

    bin_metric = BiasMetrics(binary_model)

    metrics = bin_metric.group_confusion_matrices("bias")
    metrics_flipped = bin_metric.group_confusion_matrices("bias", return_by_metric=False)

    assert "precision" in metrics.keys()
    assert len(metrics.keys()) == 8
    assert len(metrics['f1'].keys()) == 2
    assert len(metrics_flipped[1].keys()) == 8
    assert metrics_flipped[1]['accuracy_rate'] == 0.9


@responses.activate
def test_get_bias_cm_multiclass(multiclass_model):
    endpoint = "/api/v3/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasmulticlass/inferences/query"
    mock_post(endpoint, response_body=bias_query_json_result)

    mult_metric = BiasMetrics(multiclass_model)
    metrics = mult_metric.group_confusion_matrices("bias", "attr_2")
    metrics_flipped = mult_metric.group_confusion_matrices("bias", "attr_2", return_by_metric=False)

    assert "precision" in metrics.keys()
    assert len(metrics.keys()) == 8
    assert len(metrics['f1'].keys()) == 2
    assert len(metrics_flipped[1].keys()) == 8
    assert metrics_flipped[1]['accuracy_rate'] == 0.9


def test_get_bias_cm_multiclass_requires_predvalue(multiclass_model):
    mult_metric = BiasMetrics(multiclass_model)
    with pytest.raises(Exception):
        mult_metric.group_confusion_matrices("bias")


@responses.activate
def test_get_bias_pr_binary(binary_model):
    endpoint = "/api/v3/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasbinary/inferences/query"
    expected_response = {
            "query_result": [
            {
                "bias": 1,
                "positive_rate": 0.15
            },
            {
                "bias": 2,
                "positive_rate": 0.14
            }
            ],
            "sampling_threshold": 1
        }
    mock_post(endpoint, response_body=expected_response)

    bin_metric = BiasMetrics(binary_model)
    prs = bin_metric.group_positivity_rates("bias")
    prsalias = bin_metric.demographic_parity("bias")

    assert prs == prsalias    
    assert len(prs.keys()) == 2
    assert prs[2] == 0.14


def test_get_bias_pr_multiclass(multiclass_model):
    mult_metric = BiasMetrics(multiclass_model)
    with pytest.raises(Exception):
        mult_metric.group_positivity_rates("bias")